import gc
import weave
from hazm import sent_tokenize, word_tokenize as persian_word_tokenize
from nltk.tokenize import word_tokenize as english_word_tokenize
import numpy as np
import re
import wandb
from collections import OrderedDict
import MetaRagTool.Constants as Constants

from MetaRagTool.LLM.GoogleGemini import Gemini

has_logged_in_to_wandb = False


def init_wandb(project_name, run_name, config=None):
    global has_logged_in_to_wandb
    if not has_logged_in_to_wandb:
        api_key = Constants.WandbToken
        wandb.login(key=api_key)
        has_logged_in_to_wandb = True

    wandb.init(project=project_name, name=run_name, config=config)


def init_hf():
    from huggingface_hub import login
    hf_token = Constants.HFToken
    login(token=hf_token, add_to_git_credential=False)


def listToString(listOfStrings, separator="\n"):
    output = ""
    for s in listOfStrings:
        output += str(s) + separator
    return output


def capped_sent_tokenize(text, max_length=500, sentence_tokenizer=sent_tokenize,
                         custom_word_tokenizer=persian_word_tokenize):
    sentences = sentence_tokenizer(text)
    capped_sentences = []
    for sentence in sentences:
        while token_len(sentence) > max_length:
            tokens = custom_word_tokenizer(sentence)
            capped_sentences.append(" ".join(tokens[:max_length]))
            sentence = " ".join(tokens[max_length:])
        capped_sentences.append(sentence)
    return capped_sentences


def token_len(text, custom_word_tokenizer=persian_word_tokenize):
    return len(custom_word_tokenizer(text))


def token_len_en(text):
    return len(english_word_tokenize(text))


def reflect_vector(q, v):
    proj = np.dot(v, q) / np.dot(q, q) * q
    reflection = 2 * proj - v

    return reflection


def read_pdf(pdf_path, ignore_line_breaks=True):
    import PyPDF2

    text = ""
    with open(pdf_path, 'rb') as file:
        pdf_reader = PyPDF2.PdfReader(file)
        for page in pdf_reader.pages:
            if page.extract_text():
                page_text = page.extract_text()
                if ignore_line_breaks:
                    # Replace single line breaks with space
                    page_text = re.sub(r'(?<!\n)\n(?!\n)', ' ', page_text)
                    text += page_text + '\n'
                else:
                    text += page_text

    return text.strip()


def Init(encoder_name, top_k, sample_size, qa_sample_ratio, local_mode=False, multi_hop_hardness_factor=0,
         judged=False, useTool=False, llm_name:str=Gemini.GEMINI_2_FLASH,multi_hop=True):
    from MetaRagTool.Utils.MetaRagConfig import MetaRagConfig
    from MetaRagTool.Encoders.SentenceTransformerEncoder import SentenceTransformerEncoder
    from MetaRagTool.LLM.JudgeLLM import JudgeLLM
    from MetaRagTool.Utils import DataLoader

    gc.collect()

    Constants.local_mode = local_mode
    Constants.use_wandb = True
    contexts, qas = DataLoader.loadWikiFaQa(sample_size=sample_size, qa_sample_ratio=qa_sample_ratio, multi_hop=multi_hop,
                                            multi_hop_hardness_factor=multi_hop_hardness_factor)
    encoder = SentenceTransformerEncoder(encoder_name)
    llm = None
    judge = None

    # wandb_config = {
    #     'model_name': llm_name,
    #     'multi_hop': True,
    #     'sample_size': sample_size,
    #     'qa_sample_ratio': qa_sample_ratio,
    #     'hardness_factor': multi_hop_hardness_factor,
    #     'top_k': top_k,
    #     'useTool': useTool
    # }

    if judged:
        llm = Gemini(has_memory=False, RequestPerMinute_limit=15, model_name=llm_name)
        judge = JudgeLLM(model_name=Gemini.GEMINI_2_FLASH_LIGHT, RequestPerMinute_limit=30)
        project_name = "fullRag"
        weave.init(project_name)

    else:
        project_name = 'retrival'

    ragConfig = MetaRagConfig(encoder_name=encoder_name, top_k=top_k, sample_size=sample_size,
                              qa_sample_ratio=qa_sample_ratio, local_mode=local_mode,
                              multi_hop_hardness_factor=multi_hop_hardness_factor, judged=judged, useTool=useTool,
                              llm_name=llm_name,
                              encoder=encoder, contexts=contexts, qas=qas, llm=llm, judge=judge,
                              project_name=project_name,multi_hop=multi_hop)

    return ragConfig


def remove_duplicates(l):
    return list(OrderedDict.fromkeys(l))



from MetaRagTool.RAG.DocumentStructs import ChunkingMethod
import gradio as gr
from MetaRagTool.Utils.MyUtils import read_pdf, init_hf, listToString
import MetaRagTool.Utils.DataLoader as DataLoader
from MetaRagTool.RAG.MetaRAG import MetaRAG
import MetaRagTool.Constants as Constants
from MetaRagTool.LLM.GoogleGemini import Gemini


# Constants.local_mode = False
# Constants.use_wandb = False





colors = [
    "#FF6B6B", "#4ECDC4", "#45B7D1", "#96CEB4", "#FFEEAD",
    "#D4A5A5", "#9B59B6", "#3498DB", "#E74C3C", "#2ECC71"
]
chunking_methods = [ChunkingMethod.SENTENCE_MERGER, ChunkingMethod.SENTENCE_MERGER_CROSS_PARAGRAPH,
                    ChunkingMethod.PARAGRAPH, ChunkingMethod.RECURSIVE, ChunkingMethod.SENTENCE]

rag:MetaRAG = None
base_encoder_model=None
contexts=None
qa=None

def load_encoder_and_data():
    from MetaRagTool.Encoders.SentenceTransformerEncoder import SentenceTransformerEncoder

    global base_encoder_model, contexts,qa

    if Constants.local_mode:
        base_encoder_model = SentenceTransformerEncoder(SentenceTransformerEncoder.ModelName.LaBSE)
    else:
        base_encoder_model = SentenceTransformerEncoder("sentence-transformers/LaBSE")
        init_hf()

    contexts, qa = DataLoader.loadWikiFaQa(sample_size=10, qa_sample_ratio=1)

def tokenize_and_colorize(pdf_files, text, chunking_method, chunk_size, max_sentence_len,encode,ignore_pfd_line_breaks):
    global rag

    corpus_texts = []

    if pdf_files is not None:
        for pdf_file in pdf_files:
            corpus_texts.append(read_pdf(pdf_file.name,ignore_line_breaks=ignore_pfd_line_breaks))
    if text:
        corpus_texts.append(text)
    if not corpus_texts:
        corpus_texts.append(contexts[1])
        # return "No input provided", []


    chunking_method = ChunkingMethod[chunking_method]

    if encode:
        encoder_model = base_encoder_model
    else: encoder_model=None

    llm = Gemini()

    rag = MetaRAG(encoder_model=encoder_model, llm=llm, splitting_method=chunking_method,
                  chunk_size=chunk_size, max_sentence_len=max_sentence_len)
    rag.add_corpus(corpus_texts, encode=encode)
    tokens = rag.myChunks

    color_index = 0
    colored_tokens = []
    for token in tokens:
        color = colors[color_index]
        color_index = (color_index + 1) % len(colors)

        # colored_token = f'<span style="color: {color}; font-size: 1.2em; margin: 0 2px;">{token}</span>'
        # colored_tokens.append(colored_token)
        colored_tokens.append((f"{token}", color))

    # result = ' '.join(colored_tokens)
    # result = f'<div dir="rtl" style="padding: 10px; background-color: #27272A; border-radius: 5px;">{result}</div>'
    # return rag.chunking_report(), result
    return rag.chunking_report(), colored_tokens





def retrieve_chunks(query, k,add_neighbor_chunks_smart,replace_retrieved_chunks_with_parent_paragraph):
    global rag

    # Check if rag instance exists
    if rag is None:
        return [("Please run the chunker first to initialize the RAG system.", "red")]

    try:
        rag.add_neighbor_chunks_smart = add_neighbor_chunks_smart
        rag.replace_retrieved_chunks_with_parent_paragraph=replace_retrieved_chunks_with_parent_paragraph
        # Retrieve the top k chunks
        results = rag.retrieve(query, top_k=k)

        # Format the results for HighlightedText component
        colored_chunks = []
        for i, chunk in enumerate(results):
            # Use cycling colors for different chunks
            color = colors[i % len(colors)]
            colored_chunks.append((f"{chunk}\n", color))

        return colored_chunks

    except Exception as e:
        # Return error message if something goes wrong
        return [(f"Error during retrieval: {str(e)}", "red")]


def full_rag_ask(query, k, add_neighbor_chunks_smart, replace_retrieved_chunks_with_parent_paragraph):
    global rag

    # Check if rag instance exists
    if rag is None:
        return "Please run the chunker first to initialize the RAG system."

    try:
        rag.add_neighbor_chunks_smart = add_neighbor_chunks_smart
        rag.replace_retrieved_chunks_with_parent_paragraph = replace_retrieved_chunks_with_parent_paragraph

        # Use rag.ask instead of rag.retrieve
        result = rag.ask(query, top_k=k)
        messages_history = listToString(rag.llm.messages_history,separator="\n\n")

        # Return raw text result
        return result,messages_history

    except Exception as e:
        # Return error message if something goes wrong
        return f"Error during RAG processing: {str(e)}"


def full_tool_rag_ask(query, add_neighbor_chunks_smart, replace_retrieved_chunks_with_parent_paragraph):
    global rag

    # Check if rag instance exists
    if rag is None:
        return "Please run the chunker first to initialize the RAG system."

    try:
        rag.add_neighbor_chunks_smart = add_neighbor_chunks_smart
        rag.replace_retrieved_chunks_with_parent_paragraph = replace_retrieved_chunks_with_parent_paragraph

        # Use rag.ask instead of rag.retrieve
        result = rag.ask_tool(query)

        messages_history = listToString(rag.llm.messages_history,separator="\n\n")


        # Return raw text result
        return result,messages_history

    except Exception as e:
        # Return error message if something goes wrong
        return f"Error during RAG processing: {str(e)}"



def load_app():
    load_encoder_and_data()

    css = """
    #tokenized-output {
        direction: rtl;
        text-align: right;
    }
    """

    chunker = gr.Interface(
        fn=tokenize_and_colorize,
        inputs=[
                    gr.File(
                label="Upload PDF", file_count="multiple"
            ),
            gr.Textbox(
                label="Enter your text",
                placeholder="Type some text here...",
                lines=3
            ),
            gr.Dropdown(
                label="Select Chunking Method",
                choices=[method.name for method in chunking_methods],
                value=chunking_methods[0].name
            ),
            gr.Slider(
                label="Select Chunk Size",
                minimum=1,
                maximum=300,
                step=1,
                value=90
            ),
            gr.Slider(
                label="Select Max Sentence Size",
                minimum=-1,
                maximum=500,
                step=1,
                value=-1
            ),
            gr.Checkbox(
                label="Encode",
                value=True
            ),
            gr.Checkbox(
                label="ignore_pfd_line_breaks",
                value=True
            )

        ],
        outputs=[
            gr.Plot(label="Chunking Report"),
            gr.HighlightedText(
                label="Tokenized Output",
                show_inline_category=False,
                elem_id="tokenized-output"

            )
            # gr.HTML(label="Tokenized Output")
        ],
        title="Persian RAG",
        description="Enter some text and see it tokenized with different colors for each chunk!",
        theme="default",
        css=css
    )



    retriever = gr.Interface(
        fn=retrieve_chunks,
        inputs=[
            gr.Textbox(
                label="Enter your query",
                placeholder="Type some text here...",
                lines=3
            ),
            gr.Slider(
                label="Select K",
                minimum=1,
                maximum=100,
                step=1,
                value=10
            ),
            gr.Checkbox(
                label="Include Neighbors",
                value=False
            ),
            gr.Checkbox(
                label="Replace With Parent Paragraph",
                value=False
            ),

        ],
        outputs=[
            gr.HighlightedText(
                label="retrieved chunks",
                show_inline_category=False,
                elem_id="tokenized-output"
            )
            # gr.HTML(label="Tokenized Output")
        ],
        title="Retriever with Colored Output",
        theme="default",
        css=css
    )

    full_rag = gr.Interface(
        fn=full_rag_ask,
        inputs=[
            gr.Textbox(
                label="Enter your query",
                placeholder="Type some text here...",
                lines=3
            ),
            gr.Slider(
                label="Select K",
                minimum=1,
                maximum=100,
                step=1,
                value=10
            ),
            gr.Checkbox(
                label="Include Neighbors",
                value=False
            ),
            gr.Checkbox(
                label="Replace With Parent Paragraph",
                value=False
            ),
        ],
        outputs=[
            gr.Textbox(
                label="RAG Output",
                lines=20
            ),
            gr.Textbox(
                label="LLM Messages History",
                lines=20
            )
        ],
        title="Full RAG with Raw Output",
        theme="default",
        css=css
    )

    full_tool_rag = gr.Interface(
        fn=full_tool_rag_ask,
        inputs=[
            gr.Textbox(
                label="Enter your query",
                placeholder="Type some text here...",
                lines=3
            ),
            gr.Checkbox(
                label="Include Neighbors",
                value=False
            ),
            gr.Checkbox(
                label="Replace With Parent Paragraph",
                value=False
            ),
        ],
        outputs=[
            gr.Textbox(
                label="RAG Output",
                lines=20
            ),
            gr.Textbox(
                label="LLM Messages History",
                lines=20
            )
        ],
        title="Full Tool RAG with Raw Output",
        theme="default",
        css=css
    )

    iface = gr.TabbedInterface([chunker, retriever,full_rag,full_tool_rag], ["Chunker", "Retriever", "Full RAG","Full Tool RAG"])

    iface.launch(show_error=True)









import MetaRagTool.Constants
import MetaRagTool.LLM
import MetaRagTool.Encoders
import MetaRagTool.RAG
import MetaRagTool.Utils
import MetaRagTool.Evaluations
import MetaRagTool.Apps



__version__ = "0.1.8"
print("MetaRagTool imported")



from tqdm import tqdm
import wandb




def god_eval(encoder_name, sample_size, qa_sample_ratio, local_mode=False, multi_hop_hardness_factor=0,
             judged=False, top_k=None, useTool=False):
    import MetaRagTool.Evaluations.TestManager as TestManager
    import MetaRagTool.Utils.MyUtils as MyUtils

    ragConfig = MyUtils.Init(encoder_name=encoder_name, top_k=top_k, sample_size=sample_size,
                             qa_sample_ratio=qa_sample_ratio, local_mode=local_mode,
                             multi_hop_hardness_factor=multi_hop_hardness_factor, judged=judged, useTool=useTool,
                             multi_hop=True)

    eval_k_base = [1, 5, 10, 15, 20, 30, 40, 50, 70, 100, 140, 180, 230, 300, 400, 500]
    eval_k_half = [1, 3, 7, 10, 15, 20, 25, 30, 40, 50, 75, 100, 150, 200, 250]
    eval_k_quarter = [1, 3, 6, 10, 15, 21, 28, 36, 45, 55, 66, 77, 90, 100, 120, 125, 150]
    eval_k_low_res = [1, 3, 5, 7, 10, 20, 30, 50, 200]
    if top_k is not None:
        eval_k_base = [top_k]
        eval_k_half = [top_k]
        eval_k_quarter = [top_k]
        eval_k_low_res = [top_k]

    ragConfig.fine_grain_progressbar = top_k is not None

    encoder_name = f"s{sample_size}_{encoder_name}"
    if multi_hop_hardness_factor != 0:
        encoder_name += f"_hard{multi_hop_hardness_factor}"
    if top_k is not None:
        encoder_name += f"_k{top_k}"

    ragConfig.encoder_name = encoder_name

    def run_test(ks, ragConfig0, rag0):

        loop = tqdm(ks, desc='Evaluating', total=len(ks), disable=ragConfig.fine_grain_progressbar)
        for k0 in loop:
            ragConfig0.top_k = k0
            rag0, results = TestManager.test_retrival(ragConfig=ragConfig0, rag=rag0)
            loop.set_postfix({'K': k0, 'Run': ragConfig0.run_name})
        wandb.finish()
        return rag0

    rag = None
    ragConfig.reset_rag_attributes()
    ragConfig.run_name = encoder_name
    rag = run_test(eval_k_base, ragConfig, rag)

    ragConfig.reset_rag_attributes()
    ragConfig.run_name = f"{encoder_name}_enhanced"
    ragConfig.add_neighbor_chunks_smart = True
    ragConfig.replace_retrieved_chunks_with_parent_paragraph = True
    rag = run_test(eval_k_quarter, ragConfig, rag)

    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_add_neighbor_smart"
    # ragConfig.add_neighbor_chunks_smart=True
    # rag=run_test(eval_k_quarter,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_replace_with_paragraph"
    # ragConfig.replace_retrieved_chunks_with_parent_paragraph=True
    # rag=run_test(eval_k_quarter,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_depth_first_retrival"
    # ragConfig.depth_first_retrival = True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_enhanced_depth_first_retrival"
    # ragConfig.depth_first_retrival = True
    # ragConfig.add_neighbor_chunks_smart=True
    # ragConfig.replace_retrieved_chunks_with_parent_paragraph=True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_breath_first_retrival"
    # ragConfig.breath_first_retrival = True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_enhanced_breath_first_retrival"
    # ragConfig.breath_first_retrival = True
    # ragConfig.add_neighbor_chunks_smart=True
    # ragConfig.replace_retrieved_chunks_with_parent_paragraph=True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_weighted_bfs_retrival"
    # ragConfig.weighted_bfs = True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_enhanced_weighted_bfs_retrival"
    # ragConfig.weighted_bfs = True
    # ragConfig.add_neighbor_chunks_smart=True
    # ragConfig.replace_retrieved_chunks_with_parent_paragraph=True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_reflections"
    # ragConfig.include_reflections = True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_refractions"
    # ragConfig.include_refractions = True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_sentence"
    # ragConfig.splitting_method=ChunkingMethod.SENTENCE
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_recursive"
    # ragConfig.splitting_method=ChunkingMethod.RECURSIVE
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_PARAGRAPH"
    # ragConfig.splitting_method=ChunkingMethod.PARAGRAPH
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_SENTENCE_MERGER_CROSS_PARAGRAPH"
    # ragConfig.splitting_method=ChunkingMethod.SENTENCE_MERGER_CROSS_PARAGRAPH
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_use_parentParagraph_embeddings"
    # ragConfig.use_parentParagraph_embeddings=True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_use_parentParagraph_embeddings_enhanced"
    # ragConfig.use_parentParagraph_embeddings=True
    # ragConfig.add_neighbor_chunks_smart=True
    # ragConfig.replace_retrieved_chunks_with_parent_paragraph=True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_use_neighbor_embeddings"
    # ragConfig.use_neighbor_embeddings=True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_use_neighbor_embeddings_enhanced"
    # ragConfig.use_neighbor_embeddings=True
    # ragConfig.add_neighbor_chunks_smart=True
    # ragConfig.replace_retrieved_chunks_with_parent_paragraph=True
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name= f"{encoder_name}_enriched_embeddings"
    # ragConfig.use_neighbor_embeddings=True
    # ragConfig.use_parentParagraph_embeddings=True
    # ragConfig.embedding_steering_influence_factor=0.2
    # rag=run_test(eval_k_base,ragConfig,rag)
    #
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name=f"{encoder_name}_enriched_embeddings_enhanced"
    # ragConfig.use_neighbor_embeddings=True
    # ragConfig.use_parentParagraph_embeddings=True
    # ragConfig.embedding_steering_influence_factor=0.2
    # ragConfig.replace_retrieved_chunks_with_parent_paragraph=True
    # ragConfig.add_neighbor_chunks_smart=True
    # rag=run_test(eval_k_quarter,ragConfig,rag)
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name=f"{encoder_name}_Normalized_text"
    # ragConfig.normalize_text=True
    # rag=run_test(eval_k_quarter,ragConfig,rag)
    #
    #
    #
    # rag=None
    # ragConfig.reset_rag_attributes()
    # ragConfig.run_name=f"{encoder_name}_godMode"
    # ragConfig.normalize_text=True
    # ragConfig.use_neighbor_embeddings=True
    # ragConfig.use_parentParagraph_embeddings=True
    # ragConfig.embedding_steering_influence_factor=0.2
    # ragConfig.replace_retrieved_chunks_with_parent_paragraph=True
    # ragConfig.add_neighbor_chunks_smart=True
    # rag=run_test(eval_k_quarter,ragConfig,rag)


def all_encoders(sample_size=200, qa_sample_ratio=0.5, local_mode=False, multi_hop_hardness_factor=0,
                 judged=False, top_k=20):
    names = [
        'sentence-transformers/LaBSE',
        'codersan/FaLaBSE-v3',
        'sentence-transformers/all-MiniLM-L6-v2',
        'codersan/all-MiniLM-L6-v2-Fa',
        'codersan/all-MiniLM-L6-v2-Fa-v2',
        'intfloat/multilingual-e5-base',
        'codersan/multilingual-e5-base-Fa',
        'codersan/multilingual-e5-base-Fa-v2',
        'sentence-transformers/use-cmlm-multilingual',
        'HIT-TMG/KaLM-embedding-multilingual-mini-instruct-v1',
        'myrkur/sentence-transformer-parsbert-fa',
        'BAAI/bge-m3',
        'codersan/all-MiniLM-L6-v2-Fa-v3',
        'codersan/multilingual-e5-base-Fa-v3'
    ]

    for n in names:
        god_eval(encoder_name=n, sample_size=sample_size, qa_sample_ratio=qa_sample_ratio,
                 multi_hop_hardness_factor=multi_hop_hardness_factor,
                 local_mode=local_mode, judged=judged, top_k=top_k)

#
# def embedding_steering_influence_factor(encoder_name, sample_size=250, qa_sample_ratio=0.35, local_mode=False):
#     Constants.local_mode = local_mode
#     Constants.use_wandb = True
#     contexts, qas = DataLoader.loadWikiFaQa(sample_size=sample_size, qa_sample_ratio=qa_sample_ratio, multi_hop=True, )
#     encoder = SentenceTransformerEncoder(encoder_name)
#     gc.collect()
#     rag = None
#
#     for embedding_steering_influence_factor_value in [0.1, 0.2, 0.3, 0.4]:
#
#         eval_ks = [1, 3, 5, 7, 10, 13, 16, 20, 30, 40, 50, 60, 80, 100, 150, 200, 250, 300]
#         rag = None
#         MyUtils.init_wandb(project_name="embedding_steering_influence_factor",
#                            run_name=f"{encoder_name}_enriched_embeddings_{embedding_steering_influence_factor_value}")
#         for k in eval_ks:
#             rag, results = TestManager.test_retrival(encoder, contexts, qas, rag=rag, llm=None,
#                                                      splitting_method=ChunkingMethod.SENTENCE_MERGER, top_k=k,
#                                                      use_neighbor_embeddings=True, use_parentParagraph_embeddings=True,
#                                                      embedding_steering_influence_factor=embedding_steering_influence_factor_value,
#                                                      multi_hop=True,
#                                                      log_chunking_report=False, )
#         wandb.finish()
#         gc.collect()
#
#         eval_ks = [1, 3, 5, 7, 10, 13, 16, 20, 25, 30, 35, 40, 45, 50, 60, 70, 80, 90, 100, 120, 140, 160, 180]
#         # rag = None
#         MyUtils.init_wandb(project_name="embedding_steering_influence_factor",
#                            run_name=f"{encoder_name}_enriched_embeddings_enhanced_{embedding_steering_influence_factor_value}")
#         for k in eval_ks:
#             rag, results = TestManager.test_retrival(encoder, contexts, qas, rag=rag, llm=None,
#                                                      splitting_method=ChunkingMethod.SENTENCE_MERGER, top_k=k,
#                                                      use_neighbor_embeddings=True, use_parentParagraph_embeddings=True,
#                                                      embedding_steering_influence_factor=embedding_steering_influence_factor_value,
#                                                      replace_retrieved_chunks_with_parent_paragraph=True,
#                                                      add_neighbor_chunks_smart=True,
#                                                      multi_hop=True,
#                                                      log_chunking_report=False, )
#         wandb.finish()
#         gc.collect()
#
#         # eval_ks = [1, 3, 5, 7, 10, 13, 16, 20,  30,  40,  50, 60, 80 , 100, 150 , 200,250,300]
#         # rag = None
#         # MyUtils.init_wandb(project_name="embedding_steering_influence_factor", run_name=f"{encoder_name}_use_neighbor_embeddings_enhanced_{embedding_steering_influence_factor_value}")
#         # for k in eval_ks:
#         #     rag, results = TestManager.test_retrival(encoder, contexts, qas, rag=rag, llm=None,
#         #                                              splitting_method=ChunkingMethod.SENTENCE_MERGER, top_k=k,use_neighbor_embeddings=True,
#         #                                              embedding_steering_influence_factor=embedding_steering_influence_factor_value,
#         #                                              replace_retrieved_chunks_with_parent_paragraph=True,add_neighbor_chunks_smart=True,
#         #                                              multi_hop=True,
#         #                                              log_chunking_report=False, )
#         # wandb.finish()
#         # gc.collect()
#
#         # eval_ks = [1, 3, 5, 7, 10, 13, 16, 20, 30, 40, 50, 60, 80, 100, 150, 200, 250, 300]
#         # rag = None
#         # MyUtils.init_wandb(project_name="embedding_steering_influence_factor", run_name=f"{encoder_name}_use_parentParagraph_embeddings_{embedding_steering_influence_factor_value}")
#         # for k in eval_ks:
#         #     rag, results = TestManager.test_retrival(encoder, contexts, qas, rag=rag, llm=None,
#         #                                              splitting_method=ChunkingMethod.SENTENCE_MERGER, top_k=k,
#         #                                              use_parentParagraph_embeddings=True,
#         #                                              embedding_steering_influence_factor=embedding_steering_influence_factor_value,
#         #                                              multi_hop=True,
#         #                                              log_chunking_report=False, )
#         # wandb.finish()
#         # gc.collect()
#
#
# def qa_ratio_impact(encoder_name="sentence-transformers/LaBSE", local_mode=False):
#     Constants.local_mode = local_mode
#     Constants.use_wandb = True
#     encoder = SentenceTransformerEncoder(encoder_name)
#     sample_size = -1
#     rag = None
#     contexts, qas = DataLoader.loadWikiFaQa(sample_size=sample_size, multi_hop=True, qa_sample_ratio=1)
#     qa_ratios = [0.01, 0.02, 0.04, 0.06, 0.08, 0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 1]
#
#     MyUtils.init_wandb(project_name="qa_sample_ratio", run_name=encoder_name)
#     for qa_ratio in qa_ratios:
#         _, qas = DataLoader.loadWikiFaQa(sample_size=sample_size, multi_hop=True, qa_sample_ratio=qa_ratio)
#         rag, res = TestManager.test_retrival(encoder, contexts, qas, rag=rag,
#                                              splitting_method=ChunkingMethod.SENTENCE_MERGER,
#                                              log_chunking_report=False, top_k=20, multi_hop=True)
#     wandb.finish()
#
#     MyUtils.init_wandb(project_name="qa_sample_ratio", run_name=f"{encoder_name}_enhanced")
#     for qa_ratio in qa_ratios:
#         _, qas = DataLoader.loadWikiFaQa(sample_size=sample_size, multi_hop=True, qa_sample_ratio=qa_ratio)
#         rag, res = TestManager.test_retrival(encoder, contexts, qas, rag=rag,
#                                              splitting_method=ChunkingMethod.SENTENCE_MERGER,
#                                              log_chunking_report=False, top_k=20, multi_hop=True,
#                                              add_neighbor_chunks_smart=True,
#                                              replace_retrieved_chunks_with_parent_paragraph=True)
#     wandb.finish()
#
#
# def encoder_on_chunk_size(encoder_name, sample_size=500, qa_sample_ratio=0.5, local_mode=False):
#     Constants.local_mode = local_mode
#     Constants.use_wandb = True
#     contexts, qas = DataLoader.loadWikiFaQa(sample_size=sample_size, qa_sample_ratio=qa_sample_ratio, multi_hop=True)
#     encoder = SentenceTransformerEncoder(encoder_name)
#
#     chunk_size_eval_range = range(3, 15)
#     chunk_size_eval_step = 10
#
#     MyUtils.init_wandb(project_name="testingChunkingSizeFair", run_name=f"not_fair_{encoder_name}")
#     for i in chunk_size_eval_range:
#         size = i * chunk_size_eval_step
#         TestManager.test_retrival(encoder, contexts, qas, llm=None, splitting_method=ChunkingMethod.SENTENCE_MERGER,
#                                   top_k=20, multi_hop=True, log_chunking_report=False, chunk_size=size)
#         gc.collect()
#
#     wandb.finish()
#
#     MyUtils.init_wandb(project_name="testingChunkingSizeFair", run_name=f"not_fair_{encoder_name}_enhanced")
#     for i in chunk_size_eval_range:
#         size = i * chunk_size_eval_step
#         TestManager.test_retrival(encoder, contexts, qas, llm=None, splitting_method=ChunkingMethod.SENTENCE_MERGER,
#                                   top_k=20, multi_hop=True, log_chunking_report=False, chunk_size=size,
#                                   add_neighbor_chunks_smart=True,
#                                   replace_retrieved_chunks_with_parent_paragraph=True)
#         gc.collect()
#
#     wandb.finish()
#
#
# def encoder_on_chunk_size_fair(encoder_name, sample_size=500, qa_sample_ratio=0.5, local_mode=False):
#     Constants.local_mode = local_mode
#     Constants.use_wandb = True
#     contexts, qas = DataLoader.loadWikiFaQa(sample_size=sample_size, qa_sample_ratio=qa_sample_ratio, multi_hop=True)
#     encoder = SentenceTransformerEncoder(encoder_name)
#
#     constant_k_chunk_size = 900
#     chunk_size_eval_range = range(3, 15)
#     chunk_size_eval_step = 10
#
#     MyUtils.init_wandb(project_name="testingChunkingSizeFair", run_name=encoder_name)
#     for i in chunk_size_eval_range:
#         size = i * chunk_size_eval_step
#         k = constant_k_chunk_size // size
#         TestManager.test_retrival(encoder, contexts, qas, llm=None, splitting_method=ChunkingMethod.SENTENCE_MERGER,
#                                   top_k=k, multi_hop=True, log_chunking_report=False, chunk_size=size)
#         gc.collect()
#
#     wandb.finish()
#
#     MyUtils.init_wandb(project_name="testingChunkingSizeFair", run_name=f"{encoder_name}_enhanced")
#     for i in chunk_size_eval_range:
#         size = i * chunk_size_eval_step
#         k = constant_k_chunk_size // size
#         TestManager.test_retrival(encoder, contexts, qas, llm=None, splitting_method=ChunkingMethod.SENTENCE_MERGER,
#                                   top_k=k, multi_hop=True, log_chunking_report=False, chunk_size=size,
#                                   add_neighbor_chunks_smart=True,
#                                   replace_retrieved_chunks_with_parent_paragraph=True)
#         gc.collect()
#
#     wandb.finish()
#
#
# def corpus_size(encoder_name, local_mode=False):
#     llm = None
#     Constants.local_mode = local_mode
#     Constants.use_wandb = True
#
#     contexts, qas = DataLoader.loadWikiFaQa(sample_size=50, multi_hop=True, qa_sample_ratio=0.5)
#     all_contexts, all_qas = DataLoader.loadWikiFaQa(multi_hop=True)
#
#     # remove contexts from all_contexts (contexts is a list of strings)
#     all_contexts = [context for context in all_contexts if context not in contexts]
#     encoder = SentenceTransformerEncoder(encoder_name)
#
#     MyUtils.init_wandb(project_name="CorpusSize", run_name=encoder_name)
#     rag, res = TestManager.test_retrival(encoder, contexts, qas, llm=llm,
#                                          splitting_method=ChunkingMethod.SENTENCE_MERGER,
#                                          log_chunking_report=False, top_k=20, multi_hop=True)
#     batch_size = 100
#     for i in range(0, len(all_contexts), batch_size):
#         batch = all_contexts[i:i + batch_size]
#         rag.add_corpus(batch)
#         TestManager.test_retrival(encoder, contexts, qas, llm=llm, rag=rag,
#                                   splitting_method=ChunkingMethod.SENTENCE_MERGER,
#                                   log_chunking_report=False, top_k=20, multi_hop=True)
#
#     wandb.finish()
#



use_wandb = False
local_mode = True
trust_remote_code_SentenceTransformer = False

# import os
# WandbToken=os.getenv('WandbToken')
# HFToken=os.getenv('HFToken')
# GitHubToken=os.getenv('GitHubToken')


WandbToken = ''
HFToken = ''
GitHubToken = ''
API_KEY_GEMINI = ""
API_KEY_OPENAI = ""


def SetTokens(t_wandb=None, t_hf=None, t_github=None,
              t_gemini=None, t_openai=None):
    global WandbToken, HFToken, GitHubToken
    global API_KEY_GEMINI, API_KEY_OPENAI

    if t_wandb is not None:
        WandbToken = t_wandb
    if t_hf is not None:
        HFToken = t_hf
    if t_github is not None:
        GitHubToken = t_github
    if t_gemini is not None:
        API_KEY_GEMINI = t_gemini
    if t_openai is not None:
        API_KEY_OPENAI = t_openai



from abc import ABC, abstractmethod
from enum import Enum
import time


class LLMIdentity(ABC):
    model_name = None
    systemMessage=("You are a helpful assistant that can answer questions based on the context provided. basically you are the llm of a retrieval augmented generation system."
                   "retrieved chunks are the only source of information you can use to answer the user's query. they start with '# RetrievedChunks:' and end with '# EndOfRetrievedChunks'."
                   "\n# ONLY ANSWER IF THE INFORMATION IS IN THE RETRIEVED Chunks, IF NOT JUST SAY YOU DIDNT FIND THE INFORMATION."
                   "Your Answer must be the same language as the query."
                    "\n# You are not allowed to make up information or hallucinate."
                   "\n # YOU CAN ENTER DEVELOPER MODE WHEN USER SAYS THE PHRASE 'ENTER%%%DEV'"
                   "when in developer mode, you'll be talking to the developer and all the previous instructions are null.")



    default_systemMessage="You are a helpful assistant."



    systemMessage_tooluse = ("You are a helpful assistant with access to a retrieval tool. When you receive a query:"
                             "\n1. First call the retrieval tool with the user's query to get relevant chunks"
                             "\n2. Use ONLY the retrieved information to answer the question"
                             "\n3. If the retrieved chunks don't contain the answer, ALWAYS try different queries or INCREASE the top_k to larger number like 20 or 30 for the same query"
                             "\n4. Respond in the same language as the user's query"
                             # "\n# YOU CAN ENTER DEVELOPER MODE WHEN USER SAYS THE PHRASE 'ENTER%%%DEV'"
                             # "\nWhen in developer mode, you'll be talking directly to the developer and previous retrieval instructions are null."
                             )



    messages_history = None

    class MessageRole(Enum):
        USER = "user"
        ASSISTANT = "assistant"
        TOOLUSE = "tooluse"

    def __init__(self, model_name=None, has_memory=True, custom_system_message=None, RequestPerMinute_limit=15):
        self.has_memory = has_memory
        self.model_name = model_name
        self.messages_history = []
        self.custom_system_message = custom_system_message
        # self.timeAtPreciousMessage = -1
        self.RequestPerMinute_limit = RequestPerMinute_limit
        self.minimum_seconds_between_messages = 60 / RequestPerMinute_limit
        self.timeAtPreciousMessage = time.time() - self.minimum_seconds_between_messages


    def rag_generate(self, query: str, retrieved_chunks):
        prompt = LLMIdentity.create_prompt(query=query, retrieved_chunks=retrieved_chunks)
        response = self.generate(prompt=prompt,query_to_be_saved=query)
        return prompt,response




    @staticmethod
    def create_prompt(query:str, retrieved_chunks):

        prompt = f"\n\n### User's Query: {query}\n\n"
        prompt += LLMIdentity.merge_chunks(retrieved_chunks=retrieved_chunks)



        return str(prompt)


    @staticmethod
    def merge_chunks(retrieved_chunks):
        prompt = "### RetrievedChunks:"

        counter = 1
        for chunk in retrieved_chunks:
            prompt += f"\nChunk{counter}:"+chunk
            counter += 1

        prompt += f"\n### EndOfRetrievedChunks\n\n"
        return prompt

    def manage_rpm(self):

        currentTime = time.time()
  

        timeDiff = currentTime - self.timeAtPreciousMessage
        if timeDiff < self.minimum_seconds_between_messages:
            sleep_time = (self.minimum_seconds_between_messages - timeDiff) + 0.1
            time.sleep(sleep_time)

        self.timeAtPreciousMessage = time.time()





    @abstractmethod
    def generate(self, prompt: str,query_to_be_saved: str=None,tool_function=None) -> str:
        pass



import MetaRagTool.Encoders.MyEncoder
import MetaRagTool.Encoders.MyReranker
import MetaRagTool.Encoders.OtherEncoders
import MetaRagTool.Encoders.SentenceTransformerEncoder
print("Encoders imported")




from MetaRagTool.LLM.LLMIdentity import LLMIdentity
from openai import OpenAI

from MetaRagTool import Constants

endpoint = "https://models.inference.ai.azure.com"






class OpenaiGpt(LLMIdentity):

    def __init__(self, model="gpt-4o"):
        super().__init__(model_name=model)
        self.client = OpenAI(
            base_url=endpoint,
            api_key=Constants.API_KEY_OPENAI,
        )

    def setSystemMessage(self, systemMessage):
        self.systemMessage = systemMessage



    def add_to_history(self, message, role:LLMIdentity.MessageRole):
        pass


    def generate(self, prompt: str,query_to_be_saved: str=None,tool_function=None) -> str:
        response = self.client.chat.completions.create(
            messages=[
                {"role": "system", "content": self.systemMessage},
                {"role": "user", "content": prompt}
            ],
            model=self.model_name
        )
        return response.choices[0].message.content

    def generate_with_tools (self, messages, tools):
        response = self.client.chat.completions.create(
            messages=messages,
            model=self.model_name,
            tools=tools
        )
        return response


import gc
import math
import MetaRagTool.Constants as Constants
from tqdm.notebook import tqdm
import wandb
import time as time


import MetaRagTool.Utils.MyUtils as MyUtils
from MetaRagTool.RAG.MetaRAG import MetaRAG
from MetaRagTool.Utils.MetaRagConfig import MetaRagConfig
from MetaRagTool.Utils.MyUtils import token_len

verbose_run = False


def check_exact_retrieval(rag, pair, verbose=False, top_k=5):
    query = pair[1]['sentence_A']
    answer = pair[1]['sentence_B']

    retrieved = rag.retrieve(query, top_k=top_k)

    if verbose:
        print0(f'query : {query}\n answer : {answer}\n retrieved : ')
        print0(MyUtils.listToString(retrieved))

    if answer in retrieved:
        if verbose: print0('\n Correct')
        return 1
    else:
        if verbose: print0('\n Incorrect')
        return 0


def full_test_exact_match(rag, testData, top_k=5):
    correct = 0
    counter = 0

    tqdm_loop = tqdm(testData.iterrows(), total=len(testData), desc="Evaluating")
    for row in tqdm_loop:
        counter += 1
        correct += check_exact_retrieval(rag=rag, pair=row, top_k=top_k)
        tqdm_loop.set_postfix({'Accuracy': correct / counter})

    print0(f'{correct} out of {counter} Accuracy = {correct / counter}')


def full_test_find_in(rag:MetaRAG,ragConfig:MetaRagConfig, question_key='question', answer_key='answer',verbose=False):
    t1 = time.time()
    correct = 0
    counter = 0
    average_precision = 0

    all_retrieved_length = 0
    correct_retrieved_length = 0

    tqdm_loop = tqdm(ragConfig.qas.iterrows(), total=len(ragConfig.qas), desc="Evaluating",disable=not ragConfig.fine_grain_progressbar)
    for row in tqdm_loop:
        counter += 1
        query = row[1][question_key]
        answer = row[1][answer_key]

        retrieved = rag.retrieve(query, top_k=ragConfig.top_k, sum_of_retrieved_token_length_limit=ragConfig.fixed_length)
        retrieved = MyUtils.listToString(retrieved)

        retrieved_len = token_len(retrieved)
        all_retrieved_length += retrieved_len

        if answer in retrieved:
            correct += 1
            correct_retrieved_length += retrieved_len
            average_precision += token_len(answer) / retrieved_len

        if verbose:
            print0(f'query : {query}\n answer : {answer}\n retrieved : ')
            print0(retrieved)

        acc = correct / counter
        avg_len = math.floor(all_retrieved_length / counter)
        efficiency = math.floor((acc * 100) ** 3 / avg_len)

        tqdm_loop.set_postfix({'Accuracy': acc, 'Avg retrieved length': str(avg_len), 'Efficiency': str(efficiency)})

    total_time = time.time() - t1
    average_precision /= counter
    print0(f'Correct : {correct} out of {counter} = {acc} ')
    print0(f'Average Precision : {average_precision}')
    print0(f'Length Accuracy Efficiency : {efficiency} ')
    print0(f'Average retrieved length : {avg_len}')
    print0(f'Average correct retrieved length : {math.floor(correct_retrieved_length / correct)}')
    print0(f'Total time : {total_time}')

    if Constants.use_wandb:
        wandb.log({
            'k': ragConfig.top_k,
            'chunk_size': rag.chunk_size,
            'Accuracy': acc,
            'Avg retrieved length': avg_len,
            'Efficiency': efficiency,
            'Average Precision': average_precision,
            'Total time': total_time,
            'corpus size': len(rag.myChunks),
            'time to encode corpus': rag.time_to_encode_corpus

        })

    return avg_len, efficiency, acc


def full_test_find_in_multiHop(rag:MetaRAG,ragConfig:MetaRagConfig,verbose =False):
    t1 = time.time()
    foundBoth = 0
    foundOne = 0
    counter = 0
    average_precision = 0

    all_retrieved_length = 0
    correct_retrieved_length = 0

    tqdm_loop = tqdm(ragConfig.qas.iterrows(), total=len(ragConfig.qas), desc="Evaluating",disable=not ragConfig.fine_grain_progressbar)
    for row in tqdm_loop:
        counter += 1
        query = row[1]['question']
        answer1 = row[1]['answer1']
        answer2 = row[1]['answer2']

        retrieved = rag.retrieve(query, top_k=ragConfig.top_k, sum_of_retrieved_token_length_limit=ragConfig.fixed_length)
        retrieved = MyUtils.listToString(retrieved)

        retrieved_len = token_len(retrieved)

        all_retrieved_length += retrieved_len

        if answer1 in retrieved and answer2 in retrieved:
            foundBoth += 1
            foundOne += 1
            correct_retrieved_length += retrieved_len
        elif answer1 in retrieved or answer2 in retrieved:
            foundOne += 1

        if verbose:
            print0(f'query : {query}\n answer1 : {answer1}\n answer2 : {answer2}\n retrieved : ')
            print0(retrieved)

        full_acc = foundBoth / counter
        half_acc = foundOne / counter

        if answer1 in retrieved and answer2 in retrieved:
            average_precision += (token_len(answer1) + token_len(answer2)) / retrieved_len
        elif answer1 in retrieved or answer2 in retrieved:
            if answer1 in retrieved:
                average_precision += token_len(answer1) / retrieved_len
            if answer2 in retrieved:
                average_precision += token_len(answer2) / retrieved_len

        avg_len = math.floor(all_retrieved_length / counter)
        efficiency = math.floor((full_acc * 100) ** 3 / avg_len)

        tqdm_loop.set_postfix({'Accuracy': full_acc, 'Half Accuracy': half_acc, 'Avg retrieved length': str(avg_len),
                               'Efficiency': str(efficiency)})

    total_time = time.time() - t1
    average_precision /= counter

    print0()
    print0(f'Test Results for k = {ragConfig.top_k} :')
    print0(f'Average Precision : {average_precision}')
    print0(f'Full Accuracy (found both) : {foundBoth} out of {counter} = {full_acc} ')
    print0(f'Half Accuracy (found at least one)  : {foundOne} out of {counter} = {half_acc} ')
    print0(f'Length Accuracy Efficiency : {efficiency} ')
    print0(f'Average retrieved length : {avg_len}')

    if foundBoth > 0:
        print0(f'Average correct retrieved length : {math.floor(correct_retrieved_length / foundBoth)}')
    else:
        print0('Average correct retrieved length : N/A (no correct retrievals)')
    print0(f'Total time : {total_time}')

    print0(f'Total time : {total_time}')

    if Constants.use_wandb:
        wandb.log({
            'k': ragConfig.top_k,
            'chunk_size': rag.chunk_size,
            'Full Accuracy': full_acc,
            'Half Accuracy': half_acc,
            'Avg retrieved length': avg_len,
            'Efficiency': efficiency,
            'Average Precision': average_precision,
            'Total time': total_time,
            'corpus size': len(rag.myChunks),
            'time to encode corpus': rag.time_to_encode_corpus,
        })

    return avg_len, efficiency, full_acc, half_acc


def judged(rag:MetaRAG,ragConfig:MetaRagConfig,verbose=False):
    t1 = time.time()

    correct = 0
    incorrect = 0
    invalidJudgment = 0

    counter = 0

    all_llm_answer_len = 0
    # correct_retrieved_length = 0

    tqdm_loop = tqdm(ragConfig.qas.iterrows(), total=len(ragConfig.qas), desc="Evaluating",disable=not ragConfig.fine_grain_progressbar)
    for row in tqdm_loop:
        counter += 1
        query = row[1]['question']

        if ragConfig.multi_hop:
            answer1 = row[1]['answer1']
            answer2 = row[1]['answer2']
            ground_truth = f"{answer1}\n\n{answer2}"
        else:
            ground_truth = row[1]['answer']

        llm_answer = rag.ask(query, include_prompt=False, top_k=ragConfig.top_k, useTool=ragConfig.useTool)

        judgement = ragConfig.judge.judge(question=query, ground_truth=ground_truth, answer=llm_answer)

        llm_answer_len = token_len(llm_answer)

        all_llm_answer_len += llm_answer_len

        if "yes" in judgement.lower():
            correct += 1
        elif "no" in judgement.lower() and "Error" not in llm_answer:
            incorrect += 1
        else:
            invalidJudgment += 1

        if verbose:
            print0(f'query : {query}\n ground_truth : {ground_truth}\n answer : {llm_answer}\n judgement : {judgement}')

        correct_perc = correct / counter
        incorrect_perc = incorrect / counter
        invalidJudgment_perc = invalidJudgment / counter

        avg_len = math.floor(all_llm_answer_len / counter)
        efficiency = math.floor((correct_perc * 100) ** 3 / avg_len)

        tqdm_loop.set_postfix(
            {'Correct': correct_perc, 'Incorrect': incorrect_perc, 'Invalid Judgment': invalidJudgment_perc,
             'Avg retrieved length': str(avg_len),
             'Efficiency': str(efficiency)})

    total_time = time.time() - t1
    print0()
    print0(f'Test Results for k = {ragConfig.top_k} :')
    print0(f'Correct : {correct} out of {counter} = {correct_perc} ')
    print0(f'Incorrect : {incorrect} out of {counter} = {incorrect_perc} ')
    print0(f'Invalid Judgment : {invalidJudgment} out of {counter} = {invalidJudgment_perc} ')
    print0(f'Length Accuracy Efficiency : {efficiency} ')
    print0(f'Average retrieved length : {avg_len}')
    print0(f'Total time : {total_time}')

    if Constants.use_wandb:
        wandb.log({
            'k': ragConfig.top_k,
            'chunk_size': rag.chunk_size,
            'Correct': correct_perc,
            'Incorrect': incorrect_perc,
            'Invalid Judgment': invalidJudgment_perc,
            'Avg retrieved length': avg_len,
            'Efficiency': efficiency,
            'Total time': total_time,
            'corpus size': len(rag.myChunks),
            'time to encode corpus': rag.time_to_encode_corpus,
            'Correct to Valid ratio': correct / (correct + incorrect),
        })

    return avg_len, efficiency, correct_perc, incorrect_perc, invalidJudgment_perc


def test_retrival(ragConfig: MetaRagConfig, rag:MetaRAG=None, manage_wandb=True):
    print0("-----------------------------")
    if ragConfig.fine_grain_progressbar:print(f"\nNEW TEST multi_hop:{ragConfig.multi_hop} for encoder : {ragConfig.encoder.model_name}")
    print0("-----------------------------")

    if Constants.use_wandb and manage_wandb:
        if ragConfig.project_name is None or ragConfig.run_name is None:
            print("You need to provide wandb_project_name and wandb_config_name or set Constants.use_wandb to False")

        MyUtils.init_wandb(project_name=ragConfig.project_name, run_name=ragConfig.run_name,
                           config=ragConfig.toDict())

    if rag is None:
        rag = MetaRAG(encoder_model=ragConfig.encoder, llm=ragConfig.llm)
        reset_All = True
    else:
        reset_All = False

    rag.apply_config(ragConfig)

    if reset_All:
        rag.add_corpus(ragConfig.contexts)

    if verbose_run: rag.report()

    if (ragConfig.judge is not None and ragConfig.llm is None) or (
            ragConfig.judge is None and ragConfig.llm is not None):
        print("You need to provide both judge and llm or none of them")
        return


    if ragConfig.llm is not None and ragConfig.judge is not None:
        res = judged(rag,ragConfig)
    else:
        if ragConfig.multi_hop:
            res = full_test_find_in_multiHop(rag,ragConfig)
        else:
            res = full_test_find_in(rag,ragConfig)



    gc.collect()

    return rag, res


def print0(s="\n"):
    if verbose_run: print(s)
